import sys
import numpy as np
import numpy.lib.format as fmt
import numpy.testing as npt
import multiprocessing as mp
import progressbar

import click
from kspeed.kspeedcmd.kspeedcmd import pass_config
#from kspeed.kspeedcmd.commands.command_tab import *
OPTION_DICT = {
    "-u/--username": "Specify the username to login.",
    "-w/--password": "Specify the password to login.",
    "-i/--iplist": "specify the kspeed server cluster iplist.",
    "-f/--file": "Specify the object name in kspeed ",
    "-o/--objname": "Specify the object name in KSpeed.",
    "-m/--mode": "Specify the mode.",
    "-s/--shape":
    "Specify the shape of dataset to generate, e.g. \"4096, 1024\".",
    "--dtype": "Specify the dtype of dataset to generate.default='float32",
    "--datasettype": "Specify the dataset type to generate.['npy', 'hdf5']",
    "-c/--chunksize": "Specify chunk_size for dataset generating.",
    "-b/--batch_size": "Specify the batch size of getbatch from KSpeed.",
    "-p/--prefetch_num":
    "Specify the prefetching num of getbatch from KSpeed.",
    "-d/--device":
    "Specify the destinaton name to check in KSpeed.['CPU', 'GPU']",
    "-n/--process_num": "Specify number of process to check.",
}


def get_env_vars(ctx, args, incomplete):
    #return [k for k in os.environ.keys() if incomplete in k]
    preload_list = [('npy_preload', 'Preload a npy file to KSpeed.'),
                    ('file',
                     'Specify the npy file path to preload to KSpeed.'),
                    ('objname', 'Specify the object name in KSpeed.'),
                    ('mode', 'Specify the mode to preload.')]
    return [n for n in preload_list if incomplete in c[0]]


# @click.command()
# @click.argument("envvar", type=click.STRING, autocompletion=get_env_vars)
# def cmd1(envvar):
#     click.echo('Environment variable: %s' % envvar)
#    click.echo('Value: %s' % os.environ[envvar])


@click.command("npy_preload", short_help="Preload a npy file to KSpeed.")
#@click.option(autocompletion=get_env_vars)
@click.option('-f',
              '--file',
              #autocompletion=get_env_vars,
              required=True,
              type=click.Path(exists=True),
              help="Specify the npy file path to preload to KSpeed.")
@click.option('-o', '--objname', help="Specify the object name in KSpeed.")
@click.option('-m',
              '--mode',
              default='w',
              type=click.Choice(['w', 'a']),
              help="Specify the mode to preload.")
@pass_config
def npy_preload(ctx, file, objname, mode):
    """Preload a npy file to KSpeed."""
    if mode == 'a' and objname is None:
        raise ValueError('objname option could be given in mode = \'a\'.')
    ctx.kspeedcli.preloadArray(file, objname, mode)


@click.command("npy_append",
               short_help="append a npy file to a destname object in KSpeed.")
@click.option('-f',
              '--file',
              required=True,
              type=click.Path(exists=True),
              help="Specify the npy file path to append.")
@click.option('-o', '--objname', help="Specify the object name to append.")
@pass_config
def npy_append(ctx, file, objname):
    """append a npy file to a destname object in KSpeed."""
    ctx.kspeedcli.preloadArray(file, objname, 'a')


@click.command("get_npy_meta")
@click.option('-o',
              '--objname',
              required=True,
              type=str,
              help="Specify the object name in kspeed to check.")
@pass_config
def get_npy_meta(ctx, objname):
    """check the metadata of a destname object in KSpeed."""
    ctx.kspeedcli.get_npy_meta(objname)


def _npy_gen(shape, dtype, chunksize, filename):
    sample_nums = shape[0]
    shape_t = [i for i in shape]
    shape_t[0] = chunksize
    chunk_data = (np.random.random_sample(tuple(shape_t)) * 100).astype(dtype)
    with open(filename, 'wb') as f:
        header = {
            'descr': fmt.dtype_to_descr(chunk_data.dtype),
            'fortran_order': False,
            'shape': shape,
        }
        fmt.write_array_header_1_0(f, header)
        for i in range(0, sample_nums, chunksize):
            if i + chunksize > sample_nums:
                chunksize = sample_nums - i
            shape_t[0] = chunksize
            chunk_data = (np.random.random_sample(tuple(shape_t)) *
                          100).astype(dtype)
            # Write the next chunk
            f.write(chunk_data.tostring('C'))


# def _h5py_gen(sample_nums, sample_len, dtype, chunksize, filename):
#     dataset = 'ds_' + str(sample_nums) + '_' + str(sample_len)
#     with h5py.File(filename, 'a') as f:

#         # Initialize a resizable dataset to hold the output
#         sample_nums = shape[0]
#         shape_t = [i for i in shape]
#         shape_t[0] = chunksize
#         chunk_data = (np.random.random_sample(tuple(shape_t)) *
#                       100).astype(dtype)
#         maxshape = (None, ) + chunk_data.shape[1:]
#         #print(maxshape, chunk_data.shape , chunk_data.dtype)
#         dset = f.create_dataset(dataset,
#                                 shape=chunk_data.shape,
#                                 maxshape=maxshape,
#                                 chunks=chunk_data.shape,
#                                 dtype=chunk_data.dtype)
#         row_count = chunk_data.shape[0]

#         # Write the first chunk of rows
#         dset[:] = chunk_data

#         for i in range(0, sample_nums, chunksize):
#             if i + chunksize > sample_nums:
#                 chunksize = sample_nums - i
#             shape_t[0] = chunksize
#             chunk_data = (np.random.random_sample(tuple(shape_t)) *
#                           100).astype(dtype)

#             # Resize the dataset to accommodate the next chunk of rows
#             dset.resize(row_count + chunk_data.shape[0], axis=0)

#             # Write the next chunk
#             dset[row_count:] = chunk_data

#             # Increment the row count
#             row_count += chunk_data.shape[0]


@click.command("dataset_gen")
@click.option(
    '-s',
    '--shape',
    type=str,
    required=True,
    help="Specify the shape of dataset to generate, e.g. \"4096, 1024\".")
@click.option('--dtype',
              type=str,
              default='float32',
              help="Specify the dtype of dataset to generate.")
@click.option('-f',
              '--filename',
              type=str,
              help="Specify the filename of dataset to generate.")
@click.option('--datasettype',
              default='npy',
              type=click.Choice(['npy', 'hdf5']),
              help="Specify the dataset type to generate.")
@click.option('-c',
              '--chunksize',
              type=int,
              default=200,
              help="Specify chunk_size for dataset generating.")
def dataset_gen(shape, dtype, filename, datasettype, chunksize):
    """Generate dataset for test."""
    if datasettype == 'npy':
        if filename is None:
            filename = 'ds_' + '_'.join(
                shape.split(',')) + '_' + dtype + '.npy'
        click.secho(filename, fg='green', bg='black')
        shape = tuple(eval(shape))
        _npy_gen(shape, dtype, chunksize, filename)
    # elif datasettype == 'hdf5':
    #     if filename is None:
    #         filename = 'ds_' + '_'.join(shape.split(',')) + '_' + dtype + '.h5'
    #     print(filename)
    #     shape = tuple(eval(shape))
    #     _h5py_gen(shape, dtype, chunksize, filename)


@click.command("npy_check",
               short_help="check the npy file preloaded to KSpeed.")
@click.option('-f',
              '--filelist',
              required=True,
              help="Specify the host filelist to compare.")
@click.option('-o',
              '--objname',
              required=True,
              help="Specify the objname name to check in KSpeed.")
@click.option('-m',
              '--mode',
              type=click.Choice(['get', 'getbatch']),
              default='get',
              help="Specify the mode of get from KSpeed.")
@click.option('-b',
              '--batch_size',
              type=int,
              default=1,
              help="Specify the batch size of getbatch from KSpeed.")
@click.option('-p',
              '--prefetch_num',
              type=int,
              default=1,
              help="Specify the prefetching num of getbatch from KSpeed.")
@click.option('-d',
              '--device',
              type=click.Choice(['CPU', 'GPU']),
              default='CPU',
              help="Specify the destinaton name to check in KSpeed.")
@click.option('-n',
              '--process_num',
              type=int,
              default=1,
              help="Specify number of process to check.")
@pass_config
def npy_check(ctx, filelist, objname, mode, batch_size, prefetch_num, device,
              process_num):
    """check the fileset preloaded to KSpeed."""
    # mp.set_start_method('spawn')
    if device == 'GPU' and process_num > 1:
        raise ValueError('process_num option could only be 1 in GPU mode.'
                         'let process_num == 1 to enable GPU mode.')
    filelist = filelist.split(',')
    make_dataset = mp.Manager().dict()
    idrange_pre = 0
    for i in range(len(filelist)):
        data_npy = np.load(filelist[i], mmap_mode='r')
        data_npy = data_npy.reshape(-1, data_npy.shape[-1])
        id_range = len(data_npy)
        for j in range(id_range):
            make_dataset[idrange_pre + j] = [filelist[i], j]
        idrange_pre += id_range
    dataset_size = len(make_dataset)
    if mode == 'get':
        func = _npy_get_check
        assert batch_size == 1, "batch_size must be equal to 1 in get mode"
    else:
        func = _npy_getbatch_check
    if process_num == 1:
        func(0, ctx.kspeedcli, make_dataset, 0, dataset_size, objname,
             batch_size, prefetch_num, device)
        return

    idx_per_process = int(dataset_size / process_num)
    queue = mp.Queue()
    proc = []
    kspeedcli = ctx.kspeedcli
    for i in range(process_num - 1):
        p = mp.Process(target=func,
                       args=(i, kspeedcli, make_dataset, i * idx_per_process,
                             (i + 1) * idx_per_process, objname, batch_size,
                             prefetch_num, device, queue))
        p.start()
        proc.append(p)
    p = mp.Process(target=func,
                   args=(process_num - 1, kspeedcli, make_dataset,
                         (process_num - 1) * idx_per_process, dataset_size,
                         objname, batch_size, prefetch_num, device, queue))
    p.start()
    proc.append(p)

    for i in progressbar.progressbar(range(dataset_size),
                                     prefix='Checking processing:'):
        queue.get()

    for i in proc:
        i.join()


def _npy_get_check(pid,
                   kspeedcli,
                   make_dataset,
                   start,
                   end,
                   target,
                   batch_size,
                   prefetch_num,
                   device,
                   queue=None):
    if device == 'GPU':
        import torch
        cuda_device_cnt = torch.cuda.device_count()
        if cuda_device_cnt == 0:
            click.secho("No cuda device detect, use CPU", fg='red', bg='black')
            device == 'CPU'
        else:
            torch.cuda.set_device(pid % cuda_device_cnt)
    dataset = kspeedcli.loadArray(target, batch_size, prefetch_num, device)
    i = start
    bar = None
    idx_for_bar = 0
    if queue is None:
        # widgets = [
        #     progressbar.Percentage(),
        #     progressbar.Bar(), ' [',
        #     progressbar.Timer(), '] '
        # ]
        bar = progressbar.ProgressBar(  #widgets=widgets,
            prefix='Check processing:',
            max_value=end - start).start()

    while i < end:
        left_num = end - i
        current_file = make_dataset[i][0]
        start_idx_in_current_file = make_dataset[i][1]
        data_npy = np.load(current_file, mmap_mode='r')
        data_npy = data_npy.reshape(-1, data_npy.shape[-1])
        id_range = len(data_npy) - start_idx_in_current_file
        id_range_in_process = min(left_num, id_range)
        end_idx_in_current_file = start_idx_in_current_file + id_range_in_process
        for j in range(start_idx_in_current_file, end_idx_in_current_file):
            try:
                data = dataset.get(i)
                # print(data.__cuda_array_interface__)
                if device == 'CPU':
                    npt.assert_array_equal(data, data_npy[j])
                else:
                    import torch
                    src_tensor = torch.from_numpy(data_npy[j]).cuda()
                    torch.cuda.synchronize()
                    data_tensor = torch.as_tensor(data,
                                                  device=torch.device('cuda'))
                    # print(f'data_tensor.shape : {data_tensor.shape}, data_ptr : {data_tensor.data_ptr()}, device : {data_tensor.device}')
                    # print(data_tensor.__cuda_array_interface__)
                    torch.equal(data_tensor, src_tensor)
            except Exception as e:
                print(e)
                print("ERROR: kspeed load and check failed, idx=", i)
                sys.exit(-1)
            i = i + 1
            if queue is None:
                idx_for_bar = idx_for_bar + 1
                bar.update(idx_for_bar)
            else:
                queue.put(idx_for_bar)


def _npy_getbatch_check(pid,
                        kspeedcli,
                        make_dataset,
                        start,
                        end,
                        target,
                        batch_size,
                        prefetch_num,
                        device,
                        queue=None):
    if device == 'GPU':
        import torch
        cuda_device_cnt = torch.cuda.device_count()
        if cuda_device_cnt == 0:
            click.secho("No cuda device detect, use CPU", fg='red', bg='black')
            device == 'CPU'
        else:
            torch.cuda.set_device(pid % cuda_device_cnt)
    dataset = kspeedcli.loadArray(target, batch_size, prefetch_num, device)
    i = start
    bar = None
    idx_for_bar = 0
    if queue is None:
        # widgets = [
        #     progressbar.Percentage(),
        #     progressbar.Bar(), ' [',
        #     progressbar.Timer(), '] '
        # ]
        bar = progressbar.ProgressBar(  #widgets=widgets,
            prefix='Check processing:',
            max_value=end - start).start()

    while i < end:
        left_num = end - i
        current_file = make_dataset[i][0]
        start_idx_in_current_file = make_dataset[i][1]
        data_npy = np.load(current_file, mmap_mode='r')
        data_npy = data_npy.reshape(-1, data_npy.shape[-1])
        id_range = len(data_npy) - start_idx_in_current_file
        id_range_in_process = min(left_num, id_range)
        end_idx_in_current_file = start_idx_in_current_file + id_range_in_process
        for j in range(start_idx_in_current_file, end_idx_in_current_file,
                       batch_size):
            if j + batch_size > end_idx_in_current_file:
                batch_size_t = end_idx_in_current_file - j
            else:
                batch_size_t = batch_size
            try:
                data = dataset.getbatch(np.arange(i, i + batch_size_t),
                                        batch_size_t)
                # print(data.__cuda_array_interface__)
                if device == 'CPU':
                    npt.assert_array_equal(data, data_npy[j:j + batch_size_t])
                else:
                    import torch
                    src_tensor = torch.from_numpy(
                        data_npy[j:j + batch_size_t]).cuda()
                    torch.cuda.synchronize()
                    data_tensor = torch.as_tensor(data,
                                                  device=torch.device('cuda'))
                    # print(f'data_tensor.shape : {data_tensor.shape}, data_ptr : {data_tensor.data_ptr()}, device : {data_tensor.device}')
                    # print(data_tensor.__cuda_array_interface__)
                    torch.equal(data_tensor, src_tensor)
            except Exception as e:
                click.secho(e, fg='red', bg='black')
                click.secho("ERROR: kspeed load and check failed, idx=",
                            i,
                            fg='red',
                            bg='black')
                sys.exit(-1)
            if queue is None:
                idx_for_bar = idx_for_bar + batch_size_t
                bar.update(idx_for_bar)
            else:
                for h in range(i, i + batch_size_t):
                    queue.put(h)
            i = i + batch_size_t
