#!/usr/bin/env python3

# -*- coding: utf-8 -*-
import os
import sys

import click
import configparser

from kspeed.kspeedcluster import kspeedclient

CONFIGFILE = "%s/.kspeedcredentials" % os.path.expanduser('~')
CONFIGSECTION = 'KSpeedCredentials'
CONTEXT_SETTINGS = dict(auto_envvar_prefix="COMPLEX")


class Config(object):
    def __init__(self):
        self.kspeedclient = None
        config = configparser.ConfigParser()
        try:
            config.read(CONFIGFILE)
            self.username = config.get(CONFIGSECTION, 'username')
            self.password = config.get(CONFIGSECTION, 'password')
            try:
                self.iplist = config.get(CONFIGSECTION, 'iplist')
            except Exception:
                self.iplist = None 
        except Exception:
            self.username = 'admin'
            self.password = 'admin'
            self.iplist = None


pass_config = click.make_pass_decorator(Config, ensure=True)
cmd_folder = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "commands"))
namespace = {}
for filename in os.listdir(cmd_folder):
    if filename.endswith(".py") and filename.startswith("cmd_"):
        absolute_path = os.path.join(cmd_folder, filename)
        with open(absolute_path) as f:
            code = compile(f.read(), absolute_path, 'exec')
            eval(code, namespace, namespace)


class ComplexCLI(click.MultiCommand):
    def list_commands(self, ctx):
        rv = []
        for i in namespace:
            if 'Command ' in str(namespace[i]) and not str(i).startswith('_'):
                rv.append(i)
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        try:
            cmd = namespace[name]
        except Exception:
            return
        return cmd


@click.command(cls=ComplexCLI, context_settings=CONTEXT_SETTINGS)
@click.option("-u", "--username", help="Specify the username to login.")
@click.option("-w", "--password", help="Specify the password to login.")
@click.option("-i",
              "--iplist",
              help="specify the kspeed server cluster iplist.")
@pass_config
def cli(ctx, username, password, iplist):
    """A complex command line interface."""
    if username is not None:
        ctx.username = username
    if password is not None:
        ctx.password = password
    if iplist is not None:
        ctx.iplist = iplist
    try:
        if ctx.iplist is None:
            ctx.kspeedcli = kspeedclient(None, ctx.username,
                                         ctx.password)
        else:
            with open(ctx.iplist, 'r') as f:
                iplist_str = ' '.join(f.read().split())
                ctx.kspeedcli = kspeedclient(iplist_str, ctx.username,
                                            ctx.password)
    except OSError:
        click.secho('{} is not accessible.'.format(ctx.iplist),
                    fg='red',
                    bg='black')
        click.secho(
            'Please specify the correct kspeed server cluster iplist to access using -i.',
            fg='red',
            bg='black')
        sys.exit(-1)
