from image_classification.dataloaders import *
from image_classification.dataloaders import fast_collate
from image_classification.kspeedimagefolder import KSpeedImageFolder
from nvidia.dali.plugin.pytorch import DALIGenericIterator
from kspeedimagefolddali import ExternalKSpeedIterator
from redisimagefolder import RedisImageFolder
import copy

def get_kspeed_train_loader(
    data_path,
    image_size,
    batch_size,
    num_classes,
    one_hot,
    interpolation="bilinear",
    augmentation=None,
    start_epoch=0,
    workers=5,
    _worker_init_fn=None,
    prefetch_factor=2,
    memory_format=torch.contiguous_format,
):
    interpolation = {"bicubic": Image.BICUBIC, "bilinear": Image.BILINEAR}[
        interpolation
    ]
    traindir = os.path.join(data_path, "train")
    transforms_list = [
        transforms.RandomResizedCrop(image_size, interpolation=interpolation),
        transforms.RandomHorizontalFlip(),
    ]
    if augmentation == "autoaugment":
        transforms_list.append(AutoaugmentImageNetPolicy())
    iplist = []
    train_dataset = KSpeedImageFolder(
        "imagenet", "imagenet", "iplist",
        traindir,
        "imageset",
        transforms.Compose(transforms_list),
    )


    if torch.distributed.is_initialized():
        train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
    else:
        train_sampler = None


    train_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=batch_size,
        shuffle=(train_sampler is None),
        num_workers=workers,
        worker_init_fn=_worker_init_fn,
        pin_memory=True,
        sampler=train_sampler,
        collate_fn=partial(fast_collate, memory_format),
        drop_last=True,
        prefetch_factor=prefetch_factor,
        persistent_workers=True,
    )
    print(f'pid {os.getpid()} : batch_size : {batch_size},num_workers : {workers}, prefetch_factor : {prefetch_factor}, len : {len(train_loader)}')

    # return (train_loader, len(train_loader),)
    return (
        PrefetchedWrapper(train_loader, start_epoch, num_classes, one_hot),
        len(train_loader),
    )