#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kspeed.kspeedcluster import kspeedclient
import numpy as np
import multiprocessing as mp
import os

size_dict = {
    '1K': 1024,
    '2K': 2048,
    '4K': 4096,
    '8K': 8192,
    '16K': 16384,
    '32K': 32768,
    '64K': 65536,
    '128K': 131072,
    '256K': 262144,
    '512K': 524288,
    '1M': 1048576,
    '2M': 2097152,
    '4M': 4194304,
    '8M': 8388608,
    '16M': 16777216,
    '32M': 33554432,
    '64M': 67108864,
    '128M': 134217728,
    '256M': 268435456,
    '512M': 536870912,
    '1G': 1073741824,
    '2G': 2147483648,
    '4G': 4294967296,
    '8G': 8589934592,
    '16G': 17179869184,
    '32G': 34359738368,
    '64G': 68719476736,
    '128G': 137438953472,
}


def _kspeed_adduser(kspeed_iplist_str, username='admin', password='admin'):
    kspeedcli = kspeedclient(str(kspeed_iplist_str), username, password)
    kspeedcli.adduser(username, password)
    del kspeedcli


def kspeed_adduser(kspeed_iplist_str,
                   username='admin',
                   password='admin',
                   main_process=True):
    if main_process:
        _kspeed_adduser(kspeed_iplist_str, username, password)
    else:
        p = mp.Process(target=_kspeed_adduser,
                       args=(kspeed_iplist_str, username, password))
        p.start()
        p.join()


def _kspeed_deluser(username, kspeed_iplist_str):
    kspeedcli = kspeedclient(str(kspeed_iplist_str), 'admin', 'admin')
    kspeedcli.deluser(username)
    del kspeedcli


def kspeed_deluser(username, kspeed_iplist_str, main_process=True):
    if main_process:
        _kspeed_deluser(username, kspeed_iplist_str)
    else:
        p = mp.Process(target=_kspeed_deluser,
                       args=(username, kspeed_iplist_str))
        p.start()
        p.join()


def _kspeed_new_file(kspeed_iplist_str, shape, dtype, filename, username,
                     password):
    kspeedcli = kspeedclient(str(kspeed_iplist_str), username, password)
    shape_t = tuple(eval(shape))
    kspeedcli.newArray(np.array(shape_t), 2, dtype, filename)
    del kspeedcli


def kspeed_new_file(kspeed_iplist_str,
                    shape,
                    dtype,
                    filename,
                    main_process=True,
                    username='admin',
                    password='admin'):
    if main_process:
        _kspeed_new_file(kspeed_iplist_str, shape, dtype, filename, username,
                         password)
    else:
        p = mp.Process(target=_kspeed_new_file,
                       args=(kspeed_iplist_str, shape, dtype, filename,
                             username, password))
        p.start()
        p.join()


def _kspeed_cache_file(username, password, kspeed_iplist_str, filename):
    kspeedcli = kspeedclient(str(kspeed_iplist_str), username, password)
    kspeedcli.preloadArray(filename)
    del kspeedcli


def kspeed_cache_file(username,
                      password,
                      kspeed_iplist_str,
                      filename,
                      main_process=True):
    if main_process:
        _kspeed_cache_file(username, password, kspeed_iplist_str, filename)
    else:
        p = mp.Process(target=_kspeed_cache_file,
                       args=(username, password, kspeed_iplist_str, filename))
        p.start()
        p.join()


def _kspeed_rm_file(username, password, kspeed_iplist_str, filename):
    kspeedcli = kspeedclient(str(kspeed_iplist_str), username, password)
    kspeedcli.rm(filename)
    del kspeedcli


def kspeed_rm_file(username,
                   password,
                   kspeed_iplist_str,
                   filename,
                   main_process=True):
    if main_process:
        _kspeed_rm_file(username, password, kspeed_iplist_str, filename)
    else:
        p = mp.Process(target=_kspeed_rm_file,
                       args=(username, password, kspeed_iplist_str, filename))
        p.start()
        p.join()


def _get_kspeed_len(username, password, kspeed_iplist_str, filename,
                    sample_num, sample_len):
    kspeedcli = kspeedclient(str(kspeed_iplist_str), username, password)
    ksdataset = kspeedcli.loadArray(filename, 1)
    sample_len.value = ksdataset.sample_len()
    sample_num.value = ksdataset.len()
    del ksdataset, kspeedcli


def get_kspeed_len(username,
                   password,
                   kspeed_iplist_str,
                   filename,
                   main_process=True):
    if main_process:
        kspeedcli = kspeedclient(str(kspeed_iplist_str), username, password)
        ksdataset = kspeedcli.loadArray(filename, 1)
        return ksdataset.len(), ksdataset.sample_len()

    sample_num = mp.Value('i', 0)
    sample_len = mp.Value('i', 0)
    p = mp.Process(target=_get_kspeed_len,
                   args=(username, password, kspeed_iplist_str, filename,
                         sample_num, sample_len))
    p.start()
    p.join()
    return sample_num.value, sample_len.value


def parse_root(root):
    if root[-1] == '/':
        root = root[:-1]
    if root[0] != '/':
        root = os.path.abspath(os.path.join(os.getcwd(), root))
    return root


def get_top_name(root):
    """
    get top dir name
    for example:
        for root = /tmp/kspeed/imagenet-full/, return imagenet-full
    """
    root = parse_root(root)
    top_name = (root.split('/'))[-1]
    return top_name
